/**
 * @file   randsamp.c
 * @author Jade Cheng
 * @date   October 24, 2009
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define MAX(A,B) ((A) > (B) ? (A) : (B))
#define MIN(A,B) ((A) < (B) ? (A) : (B))

static int ismutation(void);
static void putline(void);
static char randactg(void);

static int          g_length;
static float        g_rate;
static const char * g_motif;
static int          g_motiflen;
static char       * g_line;

int main(int argc, char * argv[])
{
    int seed;
    int lines;
    
    if (argc != 6) {
        printf("Usage: %s motif mutation-rate lines line-length seed\n\n", argv[0]);
        printf("motif          actag...\n");
        printf("mutation-rate  0.0f ... 1.0f\n");
        printf("lines          lines of output\n");
        printf("line-length    line length and maximum motif length\n");
        printf("seed           random number generator seed\n\n");
        return EXIT_FAILURE;
    }
    
    g_motif = argv[1];
    g_motiflen = strlen(g_motif);

    if (1 != sscanf(argv[2], "%f", &g_rate) || g_rate < 0.0f || g_rate > 1.0f) {
        fprintf(stderr, "Invalid rate '%s'.\n", argv[2]);
        return EXIT_FAILURE;
    }
    
    if (1 != sscanf(argv[3], "%d", &lines) || lines <= 1) {
        fprintf(stderr, "Invalid lines '%s'.\n", argv[3]);
        return EXIT_FAILURE;
    }
    
    if (1 != sscanf(argv[4], "%d", &g_length) || g_length <= 0) {
        fprintf(stderr, "Invalid length '%s'.\n", argv[4]);
        return EXIT_FAILURE;
    }
    
    if (g_motiflen > g_length) {
        fprintf(stderr, "Motif '%s' is too long.\n", g_motif);
        return EXIT_FAILURE;
    }
    
    if (1 != sscanf(argv[5], "%d", &seed) || seed < 0) {
        fprintf(stderr, "Invalid seed '%s'.\n", argv[5]);
        return EXIT_FAILURE;
    }

    g_line = (char *)calloc(g_length + 1, 1);

    srand((unsigned)seed);
    while (lines-- > 0)
        putline();

    free(g_line);
    return EXIT_SUCCESS;
}

static int ismutation(void)
{
    return (float)(rand() % 100) < g_rate * 100.0f ? 1 : 0;
}

static void putline(void)
{
    int i, j;

    /* Fill the line with random letters. */
    for (i = 0; i < g_length; i++)
        g_line[i] = randactg();

    /* Put the motif into the line. */
    i = (g_motiflen == g_length) ? 0 : rand() % (g_length - g_motiflen);
    memcpy(g_line + i, g_motif, g_motiflen);

    /* Randomly drop a character in the motif. */
    if (0 != ismutation()) {
        int j = rand() % g_motiflen;
        memmove(g_line + i + j, g_line + i + j + 1, g_motiflen - (j + 1));
        g_line[i + (g_motiflen - 1)] = randactg();
    }

    /* Randomly insert a character into the motif. */
    if (0 != ismutation()) {
        int j = rand() % g_motiflen;
        memmove(g_line + i + j + 1, g_line + i + j, g_motiflen - j);
        g_line[i + j] = randactg();
        g_line[g_length] = '\0';
    }

    /* Randomly replace characters in the motif. */
    for (j = 0; j < g_motiflen; j++)
        if (0 != ismutation()) 
            g_line[i + j] = randactg();

    /* Randomly invert part of the motif. */
    if (0 != ismutation()) {
        int p1 = rand() % g_motiflen;
        int p2 = rand() % g_motiflen;
        int j = MIN(p1, p2);
        int k = MAX(p1, p2);
        while (j < k) {
            char ch = g_line[j];
            g_line[j] = g_line[k];
            g_line[k] = ch;
            j++, k--;
        }
    }

    printf("%s\n", g_line);
}

static char randactg(void)
{
    const char * actg = "actg";
    return actg[rand() % 4];
}
