/**
 * @file   motif.c (random)
 * @author Jade Cheng
 * @date   October 26, 2009
 */

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

static void greedymotifsearch(int * bestMotif);
static void randomswap();
static int readinput();
static int score(const int * startpos, int rowcount);
static void swap(char * a, char * b);
static void swaprows(int row1, int row2);

static const char * g_actg = "actg";
static char * samplenum;
static char * dna;
static int rows;
static int cols;
static int length;

/* -------------------------------------------------------------------------- */
/**
 * The main entry point of the program.
 *
 * @param argc The number of command line arguments.
 * @param argv The command line arguments.
 *
 * @return EXIT_SUCCESS or EXIT_FAILURE.
 */
int main(int argc, char * argv[])
{
    int        * bestMotif;
    int          i, j, seed;

    /* Check for usage errors. */
    if (argc != 5) {
        printf("Usage: %s lines line-length motif-length\n\n", argv[0]);
        printf("lines         lines of standard input\n");
        printf("line-length   length of each line of input\n");
        printf("motif-length  length of motif to find\n");
        printf("seed          random number generator seed\n\n");
        return EXIT_FAILURE;
    }
    
    /* Parse the command line input. */
    if (1 != sscanf(argv[1], "%d", &rows) || rows <= 1) {
        fprintf(stderr, "Invalid lines '%s'.\n", argv[1]);
        return EXIT_FAILURE;
    }
    if (1 != sscanf(argv[2], "%d", &cols) || cols <= 0) {
        fprintf(stderr, "Invalid line length '%s'.\n", argv[2]);
        return EXIT_FAILURE;
    }
    if (1 != sscanf(argv[3], "%d", &length) || length <= 0) {
        fprintf(stderr, "Invalid motif length '%s'.\n", argv[4]);
        return EXIT_FAILURE;
    }
    if (1 != sscanf(argv[4], "%d", &seed) || seed < 0) {
        fprintf(stderr, "Invalid seed '%s'.\n", argv[5]);
        return EXIT_FAILURE;
    }
    if (length > cols) {
        fprintf(stderr, "Motif cannot be longer than line.\n");
        return EXIT_FAILURE;
    }

    /* Allocate space for the DNA buffer and the best motif array. */
    dna = (char *)malloc(cols * rows);
    samplenum = (char *)calloc(cols, sizeof(char));
    bestMotif = (int *)calloc(rows, sizeof(int));

    /* Seed the random number generator.*/
    srand((unsigned)seed);

    /* Read the standard input. */
    if (EXIT_SUCCESS != readinput(dna, rows, cols)) {
        free(dna);
        return EXIT_FAILURE;
    }

    /* initialize the sample number array. */
    for (i = 0; i < cols; i++)
            samplenum[i] = i + 1;

    /* Search for the motif. */
    greedymotifsearch(bestMotif);

    /* Print some results. */
    printf("ID\tIndex\tMotif");

    for (i = 0; i < rows; i++) {
        printf("\n%d\t%d\t", samplenum[i], bestMotif[i]);
        for (j = 0; j < length; j++)
            fputc(g_actg[(int)dna[i * cols + bestMotif[i] + j]], stdout);
    }

    printf("\nScore:\t%d\n", score(bestMotif, rows));

    /* Release resources and return successfully. */
    free(bestMotif);
    free(samplenum);
    free(dna);
    return EXIT_SUCCESS;
}

/* -------------------------------------------------------------------------- */
/**
 * Function searches for the starting positions of the motifs that produce the
 * best score on the first two rows of the DNA matrix in a brute force manner.
 * Then updates the starting positions on the rest of the rows assuming that
 * the previous motifs have been already chosen.
 *
 * Two ways of enforcing the selection of the first two rows to be evaluated by
 * the greedy algorithm are also implemented.  One is selecting the closest
 * sequences using the hamming distances.  The other one is by a random
 * selection of two rows.
 *
 *
 * @param bestMotif The array of indexes to be filled in with the starting
 *        positions that give the best score according the greedy algorithm.
 *
 */
static void greedymotifsearch(int * bestMotif)
{
    int * startpos;
    int   end, i;

    end = cols - length;

    /* Randomly select two rows and swap them into the front of the DNA matrix. */
    randomswap();

    /* Allocate space for the aray storing the temporary starting positions. */
    startpos = (int *)calloc(rows, sizeof(int));

    /* Search for the best motifs on the first two rows in a Brute force way. */
    for (startpos[0] = 0; startpos[0] < end; startpos[0]++) {
        for (startpos[1] = 0; startpos[1] < end; startpos[1]++) {
            if (score(startpos,  2) >
                score(bestMotif, 2)) {
                bestMotif[0] = startpos[0];
                bestMotif[1] = startpos[1];
            }
        }
    }

    /* Record the starting positions of the best motifs found on the first two rows. */
    startpos[0] = bestMotif[0];
    startpos[1] = bestMotif[1];

    /* Search the rest t-2 rows assuming the previous motifs are all selected. */
    for (i = 2; i < rows; i++) {
        for (startpos[i] = 0; startpos[i] < end; startpos[i]++) {
            if (score(startpos,  i + 1) >
                score(bestMotif, i + 1))
                bestMotif[i] = startpos[i];
        }

        /* Record the starting position giving the best score on the current row. */
        startpos[i] = bestMotif[i];
    }

    free(startpos);
}

/* -------------------------------------------------------------------------- */
/**
 * Randomly choose two rows from the DNA matrix, and swap the chosen rows, row1
 * and row2, with the first and the second row in the DNA matrix respectively.
 * The updated DNA matrix's first two rows are a random selection of all rows.
 *
 */
static void randomswap()
{
    int row1, row2;

    /* Randomly select row1 and row2 from all rows. */
    row1 = rand() % rows;
    row2 = (1 + row1 + (rand() % (rows - 1))) % rows;

    /* Swap the randomly selected rows with the first two rows in the DNA matrix. */
    swaprows(row1, row2);
}

/* -------------------------------------------------------------------------- */
/**
 * Exchange the values of two bye quantities.
 *
 * @param a The address of the first byte.
 * @param b The address of the second byte.
 *
 */
static void swap(char * a, char * b)
{
    char temp = *a;
    *a = *b;
    *b = temp;
}

/* -------------------------------------------------------------------------- */
/**
 * Swap the specified rows, row1 and row2, with the first and the second row
 * respectively in the dna matrix.  At the same time, exchange the corresponding
 * elements in the sample ID array.
 *
 * @param row1 The row to be swapped with the first row in dna matrix.
 * @param row2 The row to be swapped with the second row in dan matrix.
 *
 */
static void swaprows(int row1, int row2)
{
    int i;

    assert(row1 < rows);
    assert(row2 < rows);

    for (i = 0; i < cols; i++) {
        swap(&dna[i], &dna[row1 * cols + i]);
        swap(&dna[cols + i], &dna[row2 * cols + i]);
        swap(&samplenum[0], &samplenum[row1]);
        swap(&samplenum[1], &samplenum[row2]);
    }
}

/* -------------------------------------------------------------------------- */
/**
 * Initializes the DNA buffer from row * cols of standard input.  The function
 * reads the specified number of rows of input, and for each row, it reads the
 * specified number of columns of letters from standard input.  Only 'a', 'c',
 * 'g', and 't' are allowed as input, and no whitespace is allowed except for
 * the end-of-line characters.
 *
 * @return EXIT_SUCCESS if dna is initialized successfully, or EXIT_FAILURE.
 */
static int readinput()
{
    int i, k;

    /* Loop over each row of input. */
    k = 0;
    for (i = 0; i < rows; i++) {
        int j;
        
        /* Loop over each column of input including the end-of-line. */
        for (j = 0; j <= cols; j++) {
            int ch;

            /* Read the next character and check for errors. */
            if (-1 == (ch = fgetc(stdin))) {
                fprintf(stderr, "Unexpected end of input.\n");
                return EXIT_FAILURE;
            }
            
            /* Read and store the next 'a', 'c', 't', or 'g'. */
            if (j < cols) {
                const char * ptr;

                if (NULL == (ptr = strchr(g_actg, ch))) {
                    if (ch == '\n')
                        fprintf(stderr, "Unexpected end of line.\n");
                    else
                        fprintf(stderr, "Unexpected character '%c'.\n", (char)ch);
                    return EXIT_FAILURE;
                }
                
                dna[k++] = (char)(ptr - g_actg);
                continue;
            }
            
            /* Check for an end-of-line character. */
            if (ch != '\n') {
                fprintf(stderr, "Line length mismatch.\n");
                return EXIT_FAILURE;
            }
        }
    }
    
    /* Check there is no more input. */
    if (-1 != fgetc(stdin)) {
        fprintf(stderr, "Too much data present.\n");
        return EXIT_FAILURE;
    }
    
    return EXIT_SUCCESS;
}

/* -------------------------------------------------------------------------- */
/**
 * Function computes the score for a paticular set of motifs.  It sums up the
 * max character counts on each column that starts from the specified positions
 * and ends at positions that are motif-length away from the starting positions.
 *
 * @param startpos The array of starting indexes in the DNA matrix.
 * @param rowcount The number of rows to scan to compute the score.
 *
 */
static int score(const int * startpos, int rowcount)
{
    int j;
    int total;

    /* Loop lenght many columns and sum up the scores. */
    total = 0;
    for (j = 0; j < length; j++) {
        int i;
        int max;
        int sum[4] = { 0, 0, 0, 0 };

        /* Only scan specified number of sequences. */
        for (i = 0; i < rowcount; i++) {
            char ch;

            ch = dna[i * cols + startpos[i] + j];
            sum[(int)ch]++;
        }

        /* Add the one that has the greatest count in this column. */
        max = sum[0];
        for (i = 1; i < 4; i++)
            max = sum[i] > max ? sum[i] : max;

        total += max;
    }

    return total;
}
